# Copyright 2024 IQM
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
r"""Resetting qubits to the zero state.

The reset operation is a non-unitary quantum channel that sets the state of a qubit to :math:`|0\rangle`.
"""

from __future__ import annotations

from typing import Iterable

from exa.common.data.parameter import Parameter
from exa.common.qcm_data.chip_topology import ChipTopology
from iqm.pulse.gate_implementation import (
    SINGLE_COMPONENTS_WITH_READOUT_LOCUS_MAPPING,
    CompositeGate,
    GateImplementation,
)
from iqm.pulse.timebox import TimeBox

RESET_MEASUREMENT_KEY = "__reset"
"""Constant measurement key for the measure operation required in the fast-feedback reset gate (the double underscore
emphasizes the fact that this label should not be manually used in fast feedback)."""

RESET_FEEDBACK_KEY = "__default_feedback"
"""The feedback key used in fast-feedback reset gate (the double underscore emphasizes the fact that this label should
not be manually used in fast feedback)."""


class Reset_Conditional(CompositeGate):
    r"""Conditional reset operation.

    Uses a measurement followed by a conditional PRX gate with angle :math:`\pi`.
    It is assumed the measurement projects the state into the computational basis.

    The conditional PRX implementation handles any necessary waits to accommodate for the feedback result propagation
    delay.

    This reset implementation is factorizable. It can act upon any set of locus components, and the measurement
    used in the conditional reset will be multiplexed to those components. However, only locus components that have
    readout and drive can be reset via conditional reset. Otherwise, locus components will just have their channels
    blocked.
    """

    registered_gates = ["measure"]

    def _call(self) -> TimeBox:  # type: ignore[override]
        # find locus components that are resettable via conditional reset
        resettable = tuple(
            q
            for q in self.locus
            if "drive" in self.builder.component_channels[q] and q in self.builder.chip_topology.component_to_probe_line
        )
        # try to get a qnd measurement, otherwise use default one
        try:  # TODO: should the QND measurement be its own QuantumOp instead?
            probe_timebox = self.build("measure", resettable, impl_name="constant_qnd").probe_timebox(
                RESET_MEASUREMENT_KEY, feedback_key=RESET_FEEDBACK_KEY
            )
        except (ValueError, KeyError):
            probe_timebox = self.build("measure", resettable).probe_timebox(
                RESET_MEASUREMENT_KEY, feedback_key=RESET_FEEDBACK_KEY
            )
        virtual_channels = set()
        probes = set()
        resets = []
        for component in resettable:
            virtual_channels.add(
                self.builder.get_virtual_feedback_channel_for(self.builder.get_drive_channel(component), component)
            )
            probes.add(self.builder.chip_topology.component_to_probe_line[component])
            resets.append(self.build("cc_prx", (component,))(feedback_qubit=component, feedback_key=RESET_FEEDBACK_KEY))
        # adjust the blocking neighborhood, as we know the feedforward bits are sent only to self
        # block the probes and the fast feedback virtual channels for self-consistency
        blocks = set(self.locus).union(probes).union(virtual_channels)
        # block common couplers for multi-qubit loci
        if len(self.locus) > 1:
            blocks.update(set(self.builder.chip_topology.get_connecting_couplers(self.locus)))
        probe_timebox.neighborhood_components[0] = blocks
        measure = TimeBox.composite([probe_timebox])
        measure.neighborhood_components[0] = blocks
        return TimeBox.composite([measure, resets])

    @classmethod
    def get_locus_mapping_name(cls, operation_name: str, implementation_name: str) -> str:
        return SINGLE_COMPONENTS_WITH_READOUT_LOCUS_MAPPING


class Reset_Wait(GateImplementation):
    """Reset operation by relaxation (idling for a time comparable to the relaxation time of the component).

    Adds a Wait pulse for all the (non-virtual) channels of the locus components. In addition, blocks all the probes
    associated with the locus and in case the locus is connected via couplers, blocks them as well. The operation
    is factorizable, so that the calibration data it uses (i.e. the wait duration in seconds) is defined for each
    component, and when acting on multiple components, the maximum of the associated wait durations will be applied.

    Reset by relaxation is intended to be used in the circuit initialisation between shots / segments.
    It also can be called on specific qubits inside a segment, but as it works by
    waiting longer than the qubit relaxation time, the states of all the other qubits
    will likely have been destroyed as well by the time the reset has finished.
    """

    parameters = {"duration": Parameter("", "Wait duration", "s")}

    def wait_box(self) -> TimeBox:
        """TimeBox that contains Wait instructions for all non-virtual channels associated with ``self.locus``.

        Does not block any additional components beside the locus itself.
        """
        if len(self.locus) == 1:
            waits_box = self.builder.wait(self.locus, self.calibration_data["duration"])
        else:
            prio_calibration = self.calibration_data if self.calibration_data else None
            waits_box = TimeBox.composite(
                [
                    self.builder.get_implementation(
                        self.parent.name, (q,), impl_name=self.name, priority_calibration=prio_calibration
                    ).wait_box()
                    for q in self.locus
                ]
            )
        return waits_box

    def _call(self) -> TimeBox:
        waits_box = self.wait_box()
        # block also probe lines
        blocks = {
            self.builder.chip_topology.component_to_probe_line[q]
            for q in self.locus
            if q in self.builder.chip_topology.component_to_probe_line
        }.union(waits_box.locus_components)
        # block connecting couplers for a multi-qubit locus
        if len(self.locus) > 1:
            blocks.update(set(self.builder.chip_topology.get_connecting_couplers(self.locus)))
        waits_box.neighborhood_components[0] = blocks
        # wrap into one more TimeBox for self-consistency
        return TimeBox.composite([waits_box])

    def duration_in_seconds(self) -> float:
        return max(
            self.builder.get_implementation(self.parent.name, (q,), impl_name=self.name).calibration_data["duration"]
            for q in self.locus
        )

    @classmethod
    def get_custom_locus_mapping(
        cls, chip_topology: ChipTopology, component_to_channels: dict[str, Iterable[str]]
    ) -> dict[tuple[str, ...] | frozenset[str], tuple[str, ...]] | None:
        """Supported loci: all components that have channels."""
        return {(c,): (c,) for c in chip_topology.all_components}



# -*- coding: utf-8 -*-
#
# All configuration values have a default; values that are commented out
# serve to show the default.

from __future__ import annotations

import os
import sys

from packaging.version import parse

# If extensions (or modules to document with autodoc) are in another directory,
# add these directories to sys.path here. If the directory is relative to the
# documentation root, use os.path.abspath to make it absolute, like shown here.
py_path = os.path.join(os.getcwd(), os.path.dirname(__file__), "../src")
sys.path.insert(0, os.path.abspath(py_path))


# -- Project information -----------------------------------------------------

project = "iqm.pulse"
copyright = "2019-2025, IQM Finland Oy"

# The version info for the project you're documenting, acts as replacement for
# |version| and |release|, also used in various other places throughout the
# built documents.
#
# The short X.Y version.
version = ""
# The full version, including alpha/beta/rc tags.
release = ""
try:
    from iqm.pulse import __version__ as release
except ImportError:
    pass
else:
    version = parse(release).base_version


# -- General configuration -----------------------------------------------------

# If your documentation needs a minimal Sphinx version, state it here.
needs_sphinx = "7.2"

# Add any Sphinx extension module names here, as strings. They can be extensions
# coming with Sphinx (named 'sphinx.ext.*') or your custom ones.
extensions = [
    "sphinx.ext.autodoc",
    "sphinx.ext.autosummary",
    "sphinx.ext.coverage",
    "sphinx.ext.doctest",
    "sphinx.ext.mathjax",
    "sphinx.ext.napoleon",
    "sphinx.ext.inheritance_diagram",
    "sphinx.ext.intersphinx",
    "sphinx.ext.extlinks",
    "sphinxcontrib.bibtex",
]

# Add any paths that contain templates here, relative to this directory.
templates_path = ["_templates"]

# There are two options for replacing |today|: either, you set today to some
# non-false value, then it is used:
# today = ''
# Else, today_fmt is used as the format for a strftime call.
today_fmt = "%Y-%m-%d"

# List of patterns, relative to source directory, that match files and
# directories to ignore when looking for source files.
exclude_patterns = ["_build", "_templates"]

# The name of the Pygments (syntax highlighting) style to use.
pygments_style = "sphinx"


# -- Autodoc ------------------------------------------------------------

# member ordering in autodoc output (default: 'alphabetical')
autodoc_member_order = "bysource"

# mock some imports when building the docs
# autodoc_mock_imports = ["django"]

# where should signature annotations appear in the docs, function signature or parameter description?
autodoc_typehints = "description"
# autodoc_typehints = 'description' puts the __init__ annotations into its docstring,
# which we thus have to include in the class documentation.
autoclass_content = "both"

# Sphinx 3.3+: manually clean up type alias rendering in the docs
# autodoc_type_aliases = {'TypeAlias': 'exa.experiment.somemodule.TypeAlias'}


# -- Autosummary ------------------------------------------------------------

# use autosummary to generate stub pages for API docs
autosummary_generate = True


# -- Options for HTML output ---------------------------------------------------

import sphinx_book_theme

# The theme to use for HTML and HTML Help pages.  See the documentation for
# a list of builtin themes.
html_theme = "sphinx_book_theme"

# Add any paths that contain custom themes here, relative to this directory.
html_theme_path = [sphinx_book_theme.get_html_theme_path()]

html_context = dict(display_github=False)

# Theme options are theme-specific and customize the look and feel of a theme
# further.  For a list of options available for each theme, see the
# documentation.
html_theme_options = {
    "collapse_navigation": True,
    "navigation_depth": 5,
    "use_download_button": False,
}

# A shorter title for the navigation bar.  Default is the same as html_title.
# html_short_title = None

# The name of an image file (relative to this directory) to place at the top
# of the sidebar.
html_logo = "_static/images/logo.png"

# The name of an image file (within the static path) to use as favicon of the
# docs.  This file should be a Windows icon file (.ico) being 16x16 or 32x32
# pixels large.
html_favicon = "_static/images/favicon.ico"

# Add any paths that contain custom static files (such as style sheets) here,
# relative to this directory. They are copied after the builtin static files,
# so a file named "default.css" will overwrite the builtin "default.css".
html_static_path = ["_static"]
html_css_files = ["css/custom.css"]

# If not '', a 'Last updated on:' timestamp is inserted at every page bottom,
# using the given strftime format.
html_last_updated_fmt = "%Y-%m-%d"

# If true, links to the reST sources are added to the pages.
html_show_sourcelink = False
html_copy_source = False

# Shorten class names in function type hints
python_use_unqualified_type_names = True

# Output file base name for HTML help builder.
htmlhelp_basename = "iqm-pulse-doc"


# -- MathJax options ----------------------------------------------------------

# Here we configure MathJax, mostly to define LaTeX macros.
mathjax3_config = {
    "TeX": {
        "Macros": {
            "I": r"\mathbb{I}",
            "R": r"\mathbb{R}",
            "C": r"\mathbb{C}",
            "tr": r"\text{Tr}",
            "diag": r"\text{diag}",
            "ket": [r"\left| #1 \right\rangle", 1],
            "bra": [r"\left\langle #1 \right|", 1],
        }
    }
}


# -- External mapping ------------------------------------------------------------

python_version = ".".join(map(str, sys.version_info[0:2]))
intersphinx_mapping = {
    "sphinx": ("https://www.sphinx-doc.org/en/master", None),
    "python": ("https://docs.python.org/" + python_version, None),
    "matplotlib": ("https://matplotlib.org/stable", None),
    "numpy": ("https://numpy.org/doc/stable", None),
    "scipy": ("https://docs.scipy.org/doc/scipy-1.6.3/reference", None),
    "xarray": ("https://docs.xarray.dev/en/stable", None),
    "exa.common": ("../exa-common", "../../exa-common/build/sphinx/objects.inv"),
}

use_local_target = os.getenv("USE_LOCAL_TARGET", "").lower()
if use_local_target == "true":
    intersphinx_mapping.update(
        {
            "iqm.data_definitions": (
                "../iqm-data-definitions",
                "https://iqm.gitlab-pages.iqm.fi/qccsw/iqm-data-definitions/objects.inv",
            ),
            "exa.common": (
                "../exa-common",
                "https://iqm.gitlab-pages.iqm.fi/qccsw/exa/exa-repo/exa-common/objects.inv",
            ),
        }
    )
else:
    extlinks = {
        "issue": ("https://jira.iqm.fi/browse/%s", "issue %s"),
        "mr": ("https://gitlab.iqm.fi/iqm/qccsw/exa/exa-repo/-/merge_requests/%s", "MR %s"),
    }

# -- sphinxcontrib.bibtex -------------------------------------------------

# List of all bibliography files used.
bibtex_bibfiles = ["references.bib"]



#  ********************************************************************************
#
# Copyright 2024 IQM
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
"""Utility functions."""

from __future__ import annotations

from collections.abc import Iterable
import copy
from typing import Any, get_args, get_origin

import numpy as np
from ruamel.yaml import YAML

from exa.common.data.parameter import CollectionType, DataType
from iqm.pulse.quantum_ops import QuantumOpTable


def map_waveform_param_types(type_hint: type) -> tuple[DataType, CollectionType]:
    """Map a python typehint into EXA Parameter's `(DataType, CollectionType)` tuple.

    Args:
        type: python typehint.

    Returns:
        A `(DataType, CollectionType)` tuple

    Raises:
        ValueError: for a non-supported type.

    """
    value_error = ValueError(f"Nonsupported datatype for a waveform parameter: {type_hint}")
    if hasattr(type_hint, "__iter__") and type_hint is not str:
        if type_hint == np.ndarray:
            data_type = DataType.COMPLEX  # due to np.ndarray not being generic we assume complex numbers
            collection_type = CollectionType.NDARRAY
            return (data_type, collection_type)
        if get_origin(type_hint) is list:
            collection_type = CollectionType.LIST
            type_hint = get_args(type_hint)[0]
        else:
            raise value_error
    else:
        collection_type = CollectionType.SCALAR

    if type_hint is float:
        data_type = DataType.FLOAT
    elif type_hint is int:
        data_type = DataType.INT
    elif type_hint is str:
        data_type = DataType.STRING
    elif type_hint is complex:
        data_type = DataType.COMPLEX
    elif type_hint is bool:
        data_type = DataType.BOOLEAN
    else:
        raise value_error
    return (data_type, collection_type)


def load_yaml(path: str) -> dict[str, Any]:
    """Load a YAML file from the given path, raise error if the file can't be loaded.

    Args:
        path: path to a YAML file

    Returns:
        contents of the YAML file as Python types

    """
    with open(path, encoding="utf-8") as f:
        yaml = YAML(typ="safe")  # default, if not specified, is 'rt' (round-trip)
        data = yaml.load(f)
    return data


def normalize_angle(angle: float) -> float:
    """Normalize the given angle to (-pi, pi].

    Args:
        angle: angle to normalize (in radians)

    Returns:
        ``angle`` normalized to (-pi, pi]

    """
    half_turn = np.pi
    full_turn = 2 * half_turn
    return (angle - half_turn) % -full_turn + half_turn


def phase_transformation(psi_1: float = 0.0, psi_2: float = 0.0) -> tuple[float, float]:
    r"""Implement an arbitrary (RZ, PRX, RZ) gate sequence by modifying the parameters of the
    IQ pulse implementing the PRX.

    By commutation rules we have

    .. math::
       RZ(\psi_2) \: PRX(\theta, \phi) \: RZ(\psi_1) = PRX(\theta, \phi+\psi_2) \: RZ(\psi_1 + \psi_2).

    Hence an arbitrary (RZ, PRX, RZ) gate sequence is equivalent to (RZ, PRX) with adjusted angles.

    Use case: with resonant driving, the PRX gate can be implemented using an :class:`IQPulse` instance,
    and the preceding RZ can be handled by decrementing the local oscillator phase beforehand (something
    the IQPulse instruction can also do), which is equivalent to rotating the local computational frame
    around the z axis in the opposite direction of the required quantum state rotation.

    Args:
        psi_1: RZ angle before the PRX (in rad)
        psi_2: RZ angle after the PRX (in rad)

    Returns:
        change to the PRX phase angle (in rad),
        phase increment for the IQ pulse that implements the remaining RZ (in rad)

    """
    return psi_2, -(psi_1 + psi_2)


def _validate_locus_defaults(op_name: str, definition: dict, implementations: dict) -> None:
    """Validate that locus defaults reference valid implementations.

    Args:
        op_name: Name of the gate we want to validate
        definition: Dictionary containing the parameters of the new operation
        implementations: Available implementations for the operation

    Raises:
        ValueError: If locus default references invalid implementation

    """
    for locus, impl_name in definition["defaults_for_locus"].items():
        if impl_name not in implementations:
            raise ValueError(
                f"defaults_for_locus[{locus}] implementation '{impl_name}' does not "
                f"appear in the implementations field of {op_name}."
            )


def _validate_op_attributes(op_name: str, definition: dict, op_table: QuantumOpTable) -> None:
    """Assert that the attributes of the operation (other than default implementation info) are not changed.

    Args:
        op_name: Name of the quantum operation to be added to table of quantum operations
        definition: Dictionary containing the parameters of the new operation
        op_table: Table of default or existing quantum operations

    Raises:
        ValueError: If attributes don't match defaults or are invalid

    """
    for attribute_name, value in definition.items():
        if not hasattr(op_table[op_name], attribute_name):
            raise ValueError(f"QuantumOp {op_name} has no attribute '{attribute_name}'.")

        default_value = getattr(op_table[op_name], attribute_name)
        if isinstance(value, Iterable):
            default_value = tuple(default_value)
            value = tuple(value)  # noqa: PLW2901
        if attribute_name not in ("implementations", "defaults_for_locus") and default_value != value:
            raise ValueError(
                "The QuantumOp definition provided by the user has a different value in the"
                f" attribute {attribute_name} compared to the QuantumOp definition in iqm-pulse "
                f" ({value} != {default_value}). "
                "Changing the attributes of default QuantumOps is not allowed, other than"
                "implementations and defaults_for_locus."
            )


def _process_implementations(
    op_name: str, new_implementations: dict, exposed_implementations: dict, _default_implementations: dict
) -> dict:
    """Processes implementation definitions into implementation classes.

    Args:
        op_name: Name of the quantum operation
        new_implementations: Dictionary mapping implementation names to class names
        exposed_implementations: Dictionary containing the exposed implementations
        _default_implementations: Dictionary containing the default mappings between implementations
        and implementation classes for different gates

    Returns:
        Dictionary mapping implementation names to implementation classes

    Raises:
        ValueError: If implementation class is not exposed
        ValueError: If implementation is overriding a default implementation

    """
    implementations = {}
    default_implementations = copy.deepcopy(_default_implementations).get(op_name, {})
    for new_impl_name, new_impl_cls in new_implementations.items():
        if (impl := exposed_implementations.get(new_impl_cls)) is None:
            raise ValueError(f"Implementation {new_impl_cls} has not been exposed.")
        # check if the implementation is overriding a default implementation
        for def_impl_name, def_impl_cls in default_implementations.items():
            if new_impl_name == def_impl_name:
                def_impl_class_name = def_impl_cls if isinstance(def_impl_cls, str) else def_impl_cls.__name__
                if new_impl_cls != def_impl_class_name:
                    raise ValueError(
                        f"'{def_impl_class_name}' is the default GateImplementation class for implementation '{def_impl_name}' and cannot be overridden. Consider adding a new implementation with a different name."  # noqa: E501
                    )

        implementations[new_impl_name] = impl

    return implementations



# Copyright 2024 IQM
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
"""Instructions for control instruments."""

from __future__ import annotations

import dataclasses
from dataclasses import dataclass, field
import random

from iqm.pulse.playlist.waveforms import Waveform


@dataclass(frozen=True)
class Instruction:
    """Command that can be executed by the quantum computer on a control channel.

    Has a well-specified time duration.
    """

    duration: int
    """Time duration of the instruction. In samples at the channel sample rate."""

    def __post_init__(self):
        """Add a unique id for caching purposes, which can be used if the instruction is not hashable."""
        object.__setattr__(self, "id", random.getrandbits(64))

    def validate(self) -> None:
        """Validate the instruction attributes.

        Raises:
            ValueError: something about the instruction is not ok

        """
        if self.duration < 0:
            raise ValueError(f"Instruction.duration {self.duration} is negative.")

    def copy(self, **changes) -> Instruction:
        """Make a copy of the Instruction with the given changes applied to its contents."""
        return dataclasses.replace(self, **changes)

    def get_child_instructions(self) -> tuple[Instruction, ...]:
        """Returns all the child Instructions the Instruction contains."""
        return ()

    def get_waveforms(self) -> tuple[Waveform, ...]:
        """Returns all the waveforms the Instruction contains."""
        return ()


@dataclass(frozen=True)
class Wait(Instruction):
    """Behave as if outputting zero-valued samples for the duration of the instruction.

    Used to idle QPU components. However, may be replaced with a dynamical decoupling sequence
    after the scheduling step. If you want to make sure that this does not happen, use
    :class:`Block` instead.
    """


@dataclass(frozen=True)
class Block(Instruction):
    """Behave strictly as if outputting zero-valued samples for the duration of the instruction.

    Used to block a control channel during compilation.
    A more strict version of :class:`Wait`, cannot be replaced with DD sequences during compilation.
    Converted to a :class:`.Wait` instruction at the end of compilation process.

    In "Tetris" scheduling, several Block instructions can overlap in time, whereas Waits cannot.
    """


@dataclass(frozen=True)
class VirtualRZ(Instruction):
    """Change the upconversion phase reference.

    The phase change can be done either by updating the phase of the local oscillator directly using
    a hardware instruction, or algebraically by incrementing :attr:`IQPulse.phase` of all the IQPulses
    following the VirtualRZ instruction in the :class:`.Segment`.
    """

    phase_increment: float
    """Phase increment for the local oscillator of a drive channel, in radians."""


@dataclass(frozen=True)
class RealPulse(Instruction):
    """Play a real-valued pulse."""

    wave: Waveform
    """Shape of the pulse."""
    scale: float
    """Scaling factor for the waveform."""

    def validate(self):
        super().validate()
        if abs(self.scale) > 1.0:
            raise ValueError(f"RealPulse.scale {self.scale} not in [-1, 1].")

    def get_waveforms(self) -> tuple[Waveform, ...]:
        return (self.wave,)


@dataclass(frozen=True)
class FluxPulse(RealPulse):
    """RealPulse representing a flux pulse.

    Can store RZ angles for correcting local phase shifts from the computational frame due to flux crosstalk.
    """

    rzs: tuple[tuple[str, float], ...] = field(default_factory=tuple)
    """Collection of (drive) channel names and RZ angles."""


@dataclass(frozen=True)
class IQPulse(Instruction):
    """Play an upconverted pulse that contains real in-phase and quadrature waveforms."""

    wave_i: Waveform
    """I quadrature envelope."""
    wave_q: Waveform
    """Q quadrature envelope."""
    scale_i: float = 1.0
    """Scaling factor for the I quadrature."""
    scale_q: float = 0.0
    """Scaling factor for the Q quadrature."""
    phase: float = 0.0
    """Phase of the pulse relative to the channel frequency, in radians."""
    modulation_frequency: float = 0.0
    """Modulation frequency of the waveforms, in units of the sampling rate.
    This modulation is additional to the channel frequency.
    The default value of 0.0 does not modulate.
    Note that the phase of this modulation resets for every instruction, that is, successive instances of the same
    modulated pulse are not phase coherent.
    """
    phase_increment: float = 0.0
    """Relative phase increment to the phase in the carrier frequency of this pulse and all pulses that
    are played after it. Unit: rad.
    """

    def validate(self):
        super().validate()
        if abs(self.scale_i) > 1.0:
            raise ValueError(f"IQPulse.scale_i {self.scale_i} not in [-1, 1].")
        if abs(self.scale_q) > 1.0:
            raise ValueError(f"IQPulse.scale_q {self.scale_q} not in [-1, 1].")

    def get_waveforms(self) -> tuple[Waveform, ...]:
        return (self.wave_i, self.wave_q)


@dataclass(frozen=True)
class ConditionalInstruction(Instruction):
    """Choice between multiple Instructions, depending on a condition."""

    condition: str
    """can be evaluated to an integer >= 0"""
    outcomes: tuple[Instruction, ...]
    """maps possible outcomes of the condition to the corresponding instructions"""

    def validate(self):
        super().validate()
        if not self.outcomes:
            raise ValueError("There must be at least one outcome.")
        durations = {instruction.duration for instruction in self.outcomes}
        durations.add(self.duration)
        if len(durations) != 1:
            raise ValueError(
                f"All the conditional instructions must have the same duration (now: {durations} samples)."
            )

    def get_child_instructions(self) -> tuple[Instruction, ...]:
        return self.outcomes


@dataclass(frozen=True)
class MultiplexedIQPulse(Instruction):
    """Instruction to simultaneously play multiple IQ pulses.

    Each component pulse can have an arbitrary delay from the beginning of this instruction.
    Outside the interval of the MultiplexedIQPulse, the component pulses are truncated.
    Where overlapping, samples of component pulse entries are summed.
    Where the interval of a MultiplexedIQPulse does not overlap with any of its component pulses,
    its samples are zeroes.
    """

    entries: tuple[tuple[IQPulse, int], ...]
    """(``pulse``, ``offset``) pairs.
    ``offset`` is the number of samples ``pulse`` is delayed from the beginning of the instruction.
    It has no granularity constraints. Negative values are allowed, but beginning will be truncated.
    """


@dataclass(frozen=True)
class AcquisitionMethod:
    """Describes a way to acquire readout data."""

    label: str
    """Identifier for the returned data, like ``QB1__readout.time_trace``."""
    delay_samples: int
    """Delay from beginning of probe pulse to beginning of acquisition window, in samples."""


@dataclass(frozen=True)
class TimeTrace(AcquisitionMethod):
    """Capture the raw IQ signal without integration."""

    duration_samples: int
    """Length of the capture window, in samples."""


@dataclass(frozen=True)
class ComplexIntegration(AcquisitionMethod):
    """Perform a weighted integration of the IQ raw signal, resulting in a complex number."""

    weights: IQPulse
    """Integration weights."""


@dataclass(frozen=True)
class ThresholdStateDiscrimination(ComplexIntegration):
    """Perform a weighted integration of the IQ raw signal and compare the real part of the result
    against a threshold value, resulting in a boolean.
    """

    threshold: float
    """The real part of the integration result is compared against this."""

    feedback_signal_label: str = ""
    """In fast feedback routing, the transmitted signals are associated with this label.
    ConditionalInstructions whose "condition" field has the string value of `feedback_signal_label`
    will receive the signal from this ThresholdStateDiscrimination.
    Empty string (default) means the signal is not routed anywhere.
    The same `feedback_signal_label` may not be used multiple times within the same ReadoutTrigger.
    The same `feedback_signal_label` can be used in different ReadoutTriggers and different segments.
    """


@dataclass(frozen=True)
class ReadoutTrigger(Instruction):
    """Instruction for playing a probe pulse and acquiring the associated readout results."""

    probe_pulse: MultiplexedIQPulse
    """Probe pulse to play, usually a MultiplexedIQPulse."""

    acquisitions: tuple[AcquisitionMethod, ...]
    """Active readout acquisition methods associated with this trigger instance."""

    def __add__(self, other: ReadoutTrigger) -> ReadoutTrigger:
        new_duration = max(self.duration, other.duration)
        new_probe_duration = max(self.probe_pulse.duration, other.probe_pulse.duration)
        return ReadoutTrigger(
            duration=new_duration,
            probe_pulse=MultiplexedIQPulse(
                duration=new_probe_duration, entries=self.probe_pulse.entries + other.probe_pulse.entries
            ),
            acquisitions=self.acquisitions + other.acquisitions,
        )



# Copyright 2024 IQM
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
r"""Two-qubit controlled-Z (CZ) gate.

The CZ gate flips the relative phase of the :math:`|11⟩` state.
It can be represented by the unitary matrix

.. math:: \text{CZ} = \begin{pmatrix} 1 & 0 & 0 & 0 \\ 0 & 1 & 0 & 0 \\ 0 & 0 & 1 & 0 \\ 0 & 0 & 0 & -1 \end{pmatrix}
"""

from __future__ import annotations

from dataclasses import replace
from typing import TYPE_CHECKING, Optional

import numpy as np

from exa.common.data.parameter import Parameter, Setting
from exa.common.qcm_data.chip_topology import DEFAULT_2QB_MAPPING
from iqm.pulse.gate_implementation import GateImplementation, Locus, OILCalibrationData, get_waveform_parameters
from iqm.pulse.playlist.instructions import Block, FluxPulse, Instruction, IQPulse, VirtualRZ
from iqm.pulse.playlist.schedule import Schedule
from iqm.pulse.playlist.waveforms import (
    CosineRiseFall,
    GaussianSmoothedSquare,
    ModulatedCosineRiseFall,
    Slepian,
    TruncatedGaussianSmoothedSquare,
    Waveform,
)
from iqm.pulse.timebox import TimeBox
from iqm.pulse.utils import phase_transformation

if TYPE_CHECKING:  # pragma: no cover
    from iqm.pulse.builder import ScheduleBuilder
    from iqm.pulse.quantum_ops import QuantumOp
    from iqm.pulse.timebox import TimeBox

FLUX_PULSED_QUBITS_2QB_MAPPING: str = "flux_pulsed_qubits_2qb_mapping"


class FluxPulseGate(GateImplementation):
    """Discrete two locus component gate implemented using flux pulses, virtual RZs,
    and the interaction mediated by the coupler.

    Does not take any parameters since it is discrete.

    The two locus components of the gate must be coupled by a tunable coupler.

    Consists of a flux pulse for the coupler, and possibly another one for the first locus component,
    assumed to be a qubit, both with arbitrary waveforms, and virtual RZs on both components.
    Inherit from this class and assign
    waveforms to the ``coupler_wave`` and ``qubit_wave`` pulse slots to create a specific implementation.

    Can be used as a base class for both CZ and MOVE gate implementations.

    Note: the coupler and qubit pulses typically have the same duration (given in the calibration data), and in the
    special case of the duration being zero, the gate implementation will apply ``Block(0)`` instructions
    to all the channels where it would otherwise apply flux pulses or virtual z rotations.

    Args:
        flux_pulses: mapping from flux channel name to its flux pulse
        rz: mapping from drive channel name to the virtual z rotation angle, in radians, that should be performed on it

    """

    coupler_wave: Optional[type[Waveform]]
    """Flux pulse Waveform to be played in the coupler flux AWG."""
    qubit_wave: Optional[type[Waveform]]
    """Flux pulse Waveform to be played in the qubit flux AWG."""
    root_parameters: dict[str, Parameter | Setting] = {
        "duration": Parameter("", "Gate duration", "s"),
        "rz": {
            "*": Parameter("", "Z rotation angle", "rad"),  # wildcard parameter
        },
    }
    """Parameters shared by all ``FluxPulseGate`` classes. Inheriting classes may override this if there's
    a need for additional calibration parameters."""
    excluded_parameters: list[str] = []
    """Parameters names to be excluded from ``self.parameters``. Inheriting classes may override this if certain
    parameters are not wanted in that class (also parameters defined by the waveforms can be excluded)."""

    def __init__(
        self,
        parent: QuantumOp,
        name: str,
        locus: Locus,
        calibration_data: OILCalibrationData,
        builder: ScheduleBuilder,
    ):
        super().__init__(parent, name, locus, calibration_data, builder)
        duration = calibration_data["duration"]
        flux_pulses = {}

        def build_flux_pulse(waveform_class: type[Waveform], component_name: str, cal_node_name: str) -> None:
            """Uses a part of the gate calibration data to prepare a flux pulse for the given component."""
            flux_channel = builder.get_flux_channel(component_name)
            params = self.convert_calibration_data(
                calibration_data[cal_node_name],
                self.parameters[cal_node_name],
                builder.channels[flux_channel],
                duration,
            )
            # TODO convert_calibration_data should be able to do this too
            n_samples = builder.channels[flux_channel].duration_to_int_samples(duration) if duration > 0 else 0
            params["n_samples"] = n_samples
            amplitude = params.pop("amplitude")
            flux_pulses[flux_channel] = FluxPulse(
                duration=n_samples,
                wave=waveform_class(**params),
                scale=amplitude,
            )

        if self.coupler_wave is not None:
            build_flux_pulse(self.coupler_wave, builder.chip_topology.get_coupler_for(*locus), "coupler")

        if self.qubit_wave is not None:
            # the pulsed qubit is always the first one of the locus
            build_flux_pulse(self.qubit_wave, locus[0], "qubit")

        rz = calibration_data["rz"]
        # rz angles are required for the two locus components, and are optional for every other drivable QPU component
        # NOTE computational resonators cannot be driven directly, instead they have "virtual" drive channels
        # that are removed by the compiler at the end of the compilation, and implemented by other means.
        for c in locus:
            if c not in rz:
                # TODO self.qualified_name when __init__ and construct are merged
                raise ValueError(
                    f"{parent.name}.{name}: {locus}: Calibration is missing an RZ angle for locus component {c}."
                )
        rz_locus = {builder.get_drive_channel(c): angle for c, angle in rz.items() if c in locus}
        rz_not_locus = tuple((builder.get_drive_channel(c), angle) for c, angle in rz.items() if c not in locus)
        # No driving must happen on any of the affected components during the flux pulses,
        # hence the virtual z rotations must use up their entire duration.
        T = max(pulse.duration for pulse in flux_pulses.values())
        # The gate takes no parameters, so we may build and cache the entire Schedule here.
        schedule: dict[str, list[Instruction]] = {}
        for channel, angle in rz_locus.items():
            # the virtual rz technique requires decrementing the drive phase by the rz angle
            schedule[channel] = [VirtualRZ(duration=T, phase_increment=-angle)]
        vzs_inserted = False  # insert the long-distance Vzs to the first flux pulse (whatever that is)
        for channel, flux_pulse in flux_pulses.items():
            if rz_not_locus and not vzs_inserted:
                schedule[channel] = [replace(flux_pulse, rzs=rz_not_locus)]
                vzs_inserted = True
            else:
                schedule[channel] = [flux_pulse]
        affected_components = set(locus)
        affected_components.add(builder.chip_topology.get_coupler_for(*locus))
        self._affected_components = affected_components
        self._schedule = Schedule(schedule if T > 0 else {c: [Block(0)] for c in schedule}, duration=T)

    def __init_subclass__(
        cls, /, coupler_wave: Optional[type[Waveform]] = None, qubit_wave: Optional[type[Waveform]] = None
    ):
        """Store the Waveform types used by this subclass, and their parameters.

        NOTE: if ``MyCZ`` is a subclass of ``FluxPulseGate``, with some defined coupler and qubit waves, further
        inheriting from it like this ``class MySubSubClass(MyCZ, coupler_wave=Something, qubit_wave=SomethingElse)``
        changes the waves accordingly. If you do not provide any waves: ``class MySubSubClass(MyCZ)``,
        the waves defined in ``MyCZ`` will be retained. If you provide just one wave:
        ``class MySubSubClass(MyCZ, coupler_wave=Something)``, the other wave will be initialised as ``None``.

        Args:
            coupler_wave: flux pulse waveform to be played in the coupler flux AWG. Can be set as `None` if
                no coupler flux pulse should be played in this gate implementation.
            qubit_wave: flux pulse waveform to be played in the qubit flux AWG. Can be set as `None` if
                no qubit flux pulse should be played in this gate implementation.

        """
        # fix __init_subclass__ behaviour for further inheritance from a subclass of FluxPulseGate
        # we can skip this function if the class attributes are already stored in the parent class
        # and the subsubclass definition does not change these
        # see more info in: https://stackoverflow.com/questions/55183288/inheriting-init-subclass-parameters
        # the unintuitive default ``None`` values and handling of these values is for overcoming this issue
        # so that the method itself behaves as expected in successive subclassing
        if coupler_wave is None and qubit_wave is None and hasattr(cls, "coupler_wave") and hasattr(cls, "qubit_wave"):
            return
        cls.coupler_wave = coupler_wave
        cls.qubit_wave = qubit_wave
        cls.symmetric = cls.qubit_wave is None

        root_parameters = {k: v for k, v in cls.root_parameters.items() if k not in cls.excluded_parameters}
        parameters = {}
        if coupler_wave is not None:
            parameters["coupler"] = get_waveform_parameters(coupler_wave, label_prefix="Coupler flux pulse ")
            parameters["coupler"]["amplitude"] = Parameter("", "Coupler flux pulse amplitude", "")

        if qubit_wave is not None:
            parameters["qubit"] = get_waveform_parameters(qubit_wave, label_prefix="Qubit flux pulse ")
            parameters["qubit"]["amplitude"] = Parameter("", "Qubit flux pulse amplitude", "")

        cls.parameters = root_parameters | {k: v for k, v in parameters.items() if k not in cls.excluded_parameters}

    def _call(self) -> TimeBox:
        timebox = self.to_timebox(self._schedule)
        timebox.neighborhood_components[0] = self._affected_components
        return timebox

    def duration_in_seconds(self) -> float:
        if self._schedule.duration == 0:
            return 0.0
        return self.builder.channels[list(self._schedule.channels())[0]].duration_to_seconds(self._schedule.duration)

    @classmethod
    def get_locus_mapping_name(cls, operation_name: str, implementation_name: str) -> str:
        if cls.qubit_wave:
            return FLUX_PULSED_QUBITS_2QB_MAPPING
        return DEFAULT_2QB_MAPPING


class CZ_GaussianSmoothedSquare(FluxPulseGate, coupler_wave=GaussianSmoothedSquare):
    """CZ gate using a GaussianSmoothedSquare flux pulse on the coupler."""


class CZ_Slepian(FluxPulseGate, coupler_wave=Slepian):
    """CZ gate using a Slepian flux pulse on the coupler."""


class CZ_TruncatedGaussianSmoothedSquare(FluxPulseGate, coupler_wave=TruncatedGaussianSmoothedSquare):
    """CZ gate using a TruncatedGaussianSmoothedSquare flux pulse on the coupler."""


class CZ_Slepian_CRF(FluxPulseGate, coupler_wave=Slepian, qubit_wave=CosineRiseFall):
    """CZ gate using a Slepian flux pulse on the coupler and a CosineRiseFall flux pulse on the qubit."""


class CZ_CRF(FluxPulseGate, coupler_wave=CosineRiseFall):
    """CZ gate using a CosineRiseFall flux pulse on the coupler."""


class FluxPulseGate_TGSS_CRF(FluxPulseGate, coupler_wave=TruncatedGaussianSmoothedSquare, qubit_wave=CosineRiseFall):
    """CZ gate using a TruncatedGaussianSmoothedSquare flux pulse on the coupler and a CosineRiseFall
    flux pulse on the qubit.
    """


class FluxPulseGate_CRF_CRF(FluxPulseGate, coupler_wave=CosineRiseFall, qubit_wave=CosineRiseFall):
    """CZ gate using a CosineRiseFall flux pulse on the coupler and on the qubit."""


AC_STARK_PULSED_QUBITS_2QB_MAPPING: str = "ac_stark_pulsed_qubits_2qb_mapping"


class CouplerFluxPulseQubitACStarkPulseGate(GateImplementation):
    """Base class for CZ gates with coupler flux pulse and a qubit AC Stark pulse.

    Analogous to the fast qubit flux pulse, the AC Stark pulse can tune the frequency of the qubit. Together with the
    coupler flux pulse, this can implement a fast qubit pulsed CZ gate.

    """

    coupler_wave: Optional[type[Waveform]]
    """Flux pulse Waveform to be played in the coupler flux AWG."""
    qubit_drive_wave: Optional[type[Waveform]]
    """Qubit drive pulse waveform to be played in the qubit drive AWG."""

    root_parameters: dict[str, Parameter | Setting] = {
        "duration": Parameter("", "Gate duration", "s"),
        "rz": {
            "*": Parameter("", "Z rotation angle", "rad"),
        },
    }
    excluded_parameters: list[str] = []
    """Parameters names to be excluded from ``self.parameters``. Inheriting classes may override this if certain
    parameters are not wanted in that class (also parameters defined by the waveforms can be excluded)."""

    def __init__(
        self,
        parent: QuantumOp,
        name: str,
        locus: Locus,
        calibration_data: OILCalibrationData,
        builder: ScheduleBuilder,
    ):
        super().__init__(parent, name, locus, calibration_data, builder)

        duration = calibration_data["duration"]
        flux_pulses = {}
        qubit_drive_pulses = {}
        rz = calibration_data["rz"]

        def build_flux_pulse(waveform_class: type[Waveform], component_name: str, cal_node_name: str) -> None:
            """Uses a part of the gate calibration data to prepare a flux pulse for the given component."""
            flux_channel = builder.get_flux_channel(component_name)
            params = self.convert_calibration_data(
                calibration_data[cal_node_name],
                self.parameters[cal_node_name],
                builder.channels[flux_channel],
                duration,
            )
            n_samples = builder.channels[flux_channel].duration_to_int_samples(duration) if duration > 0 else 0
            params["n_samples"] = n_samples
            amplitude = params.pop("amplitude")
            flux_pulses[flux_channel] = FluxPulse(
                duration=n_samples,
                wave=waveform_class(**params),
                scale=amplitude,
            )

        def build_ac_stark_pulse(component_name: str, cal_node_name: str) -> None:
            """Uses a part of the gate calibration data to prepare a flux pulse for the given component."""
            drive_channel = builder.get_drive_channel(component_name)
            params = self.convert_calibration_data(
                calibration_data[cal_node_name],
                self.parameters[cal_node_name],
                builder.channels[drive_channel],
                duration,
            )
            n_samples = builder.channels[drive_channel].duration_to_int_samples(duration) if duration > 0 else 0
            params["n_samples"] = n_samples
            params["phase_increment"] = rz[component_name]
            qubit_drive_pulses[drive_channel] = self._ac_stark_pulse(**params)

        if self.coupler_wave is not None:
            build_flux_pulse(self.coupler_wave, builder.chip_topology.get_coupler_for(*locus), "coupler")

        if self.qubit_drive_wave is not None:
            # the pulsed qubit is always the first one of the locus
            build_ac_stark_pulse(locus[0], "first_qubit")
            build_ac_stark_pulse(locus[1], "second_qubit")

        T = max(pulse.duration for pulse in list(flux_pulses.values()) + list(qubit_drive_pulses.values()))
        schedule: dict[str, list[Instruction]] = {}

        for channel, qubit_drive_pulse in qubit_drive_pulses.items():
            schedule[channel] = [qubit_drive_pulse]
        rz_not_locus = tuple((builder.get_drive_channel(c), angle) for c, angle in rz.items() if c not in locus)
        for channel, flux_pulse in flux_pulses.items():  # just one flux pulse here
            if rz_not_locus:
                schedule[channel] = [replace(flux_pulse, rzs=rz_not_locus)]
            else:
                schedule[channel] = [flux_pulse]

        affected_components = set(locus)
        affected_components.add(builder.chip_topology.get_coupler_for(*locus))
        self._affected_components = affected_components

        self._schedule = Schedule(schedule) if T > 0 else Schedule({c: [Block(0)] for c in schedule})

    def __init_subclass__(
        cls, /, coupler_wave: Optional[type[Waveform]] = None, qubit_drive_wave: Optional[type[Waveform]] = None
    ):
        """Store the Waveform types used by this subclass, and their parameters."""
        cls.coupler_wave = coupler_wave
        cls.qubit_drive_wave = qubit_drive_wave
        cls.symmetric = True

        root_parameters = {k: v for k, v in cls.root_parameters.items() if k not in cls.excluded_parameters}
        parameters = {}
        if coupler_wave is not None:
            parameters["coupler"] = get_waveform_parameters(coupler_wave)
            parameters["coupler"]["amplitude"] = Parameter("", "amplitude", "")

        if qubit_drive_wave is not None:
            for cal_node_name in ["first_qubit", "second_qubit"]:
                # Same AC Stark pulse waveform for both qubits
                parameters[cal_node_name] = get_waveform_parameters(qubit_drive_wave)
                parameters[cal_node_name]["amplitude"] = Parameter("", "amplitude", "")

        cls.parameters = root_parameters | {k: v for k, v in parameters.items() if k not in cls.excluded_parameters}

    @classmethod
    def _ac_stark_pulse(
        cls,
        *,
        phase: float,
        amplitude: float,
        phase_increment: float,
        **kwargs,
    ) -> IQPulse:
        """Returns an AC Stark pulse which consists of a modulated I and modulated Q waveform, where the Q quadrature
        has an additional phase of -pi/2.
        """
        _, phase_increment = phase_transformation(0.0, phase_increment)

        wave_i = cls.qubit_drive_wave(phase=phase, **kwargs)  # type: ignore
        wave_q = cls.qubit_drive_wave(phase=phase - np.pi / 2, **kwargs)  # type: ignore
        return IQPulse(
            kwargs["n_samples"],
            wave_i=wave_i,
            wave_q=wave_q,
            scale_i=amplitude,
            scale_q=amplitude,
            phase_increment=phase_increment,
        )

    def _call(self) -> TimeBox:
        timebox = self.to_timebox(self._schedule)
        timebox.neighborhood_components[0] = self._affected_components
        return timebox

    def duration_in_seconds(self) -> float:
        if self._schedule.duration == 0:
            return 0.0
        return self.builder.channels[list(self._schedule.channels())[0]].duration_to_seconds(self._schedule.duration)

    @classmethod
    def get_locus_mapping_name(cls, operation_name: str, implementation_name: str) -> str:
        if cls.qubit_drive_wave:
            return AC_STARK_PULSED_QUBITS_2QB_MAPPING
        return DEFAULT_2QB_MAPPING


class CZ_Slepian_ACStarkCRF(
    CouplerFluxPulseQubitACStarkPulseGate,
    coupler_wave=Slepian,
    qubit_drive_wave=ModulatedCosineRiseFall,
):
    """Controlled-Z two-qubit gate.

    CZ gate implemented using a slepian flux pulse for the coupler and a modulated cosine rise fall (CRF) AC Stark
    pulse on one qubit.
    """


class CZ_CRF_ACStarkCRF(
    CouplerFluxPulseQubitACStarkPulseGate,
    coupler_wave=CosineRiseFall,
    qubit_drive_wave=ModulatedCosineRiseFall,
):
    """Controlled-Z two-qubit gate.

    CZ gate implemented using a cosine rise fall flux pulse for the coupler and a modulated
    cosine rise fall (CRF) AC Stark pulse on one qubit.
    """



#  ********************************************************************************
#  Copyright (c) 2019-2024 IQM Finland Oy.
#  All rights reserved. Confidential and proprietary.
#
#  Distribution or reproduction of any information contained herein
#  is prohibited without IQM Finland Oy’s prior written permission.
#  ********************************************************************************
"""Common setup for subdirectories."""

from pathlib import Path

from setuptools import setup


def get_version():
    """Get version from the file version.txt."""
    build_version_fpath = Path("version.txt")
    if build_version_fpath.exists():
        # Assume there's only one line specifying the version in the file
        return build_version_fpath.read_text().strip()
    return "0.1dev0"


if __name__ == "__main__":
    setup(version=get_version())
